%%%
%%% stephen potter july '03
%%% a first try at a generic 'transformation' broker
%%% 

%%%
%%% The brokering algorithm is as follows:
%%% 1. Let the solution set, S={}. Let the possibles set, P={}. Search
%%% the advertised services to find possible 'terminal' services for
%%% the current query (services that *may* constitute the final element 
%%% in a sequence of services). Add each of these to P to represent the
%%% final element of a potential service sequence.
%%% 
%%% 2. If P={} then stop. Return S as the solution to this query.
%%%
%%% 3. Consider each element p of P in turn to determine whether it 
%%% constitutes a 'complete' service sequence for the query: if the
%%% input knowledge of the query and of the first element of p match
%%% then this is a complete sequence and p is transferred from P to S.
%%%
%%% 4. For each incomplete sequence p in P search the current services
%%% for any other known service s that could plausibly precede these in
%%% a service sequence: s can plausibly precede p is the output
%%% knowledge of s matches the input knowledge of p. For each s that 
%%% meets this criterion, a new plausible service sequence consisting
%%% of s appended to the front of p is added to P. Finally, remove p
%%% from P.
%%%
%%% 5. Repeat from step 2.

%%%
%%% Representations:
%%%
%%% A service is represented by the tuple:
%%%   service(URI, Name, Ontology, Lifecycle, Transformation)
%%% where
%%%   * URI is the URI of the service in question.
%%%   * Name is the name of this service.
%%%   * Ontology is the (name of the prolog file containing the) transformation 
%%% ontology used to describe the Transformation.
%%%   * Lifecycle is the lifecycle itself, in the form:
%%%      trans(input,output)=>[terms+]
%%%      if input, output are not ground, it is expected that the variables
%%%   will appear in the (Prolog) terms (effectively used to define the 
%%%   constraints on input and output and the transformation from one to the
%%%   other). Note that terms that refer (perhaps indirectly) to input concern
%%%   the input knowledge, those refering to output concern the output 
%%%   knowledge, and those refering to both concern the transformation.
%%%   * Transformation is the transformation implied within the lifecycle: this
%%%   is the tuple transformation(input,output,transformation). Each of the 
%%%   elements input and output is a list of lists, each of which represents the
%%%   description of one body of knowledge. 
%%% 
%%% A service query will have the same form. 
:- use_module(library(lists)).
:- op(800, xfx, [=>]).
:- dynamic ontology/2, believed/2, service/4.

%%% Ontology->Load: loads life-cycle ontology from file
loadOntology(File) :-
	%retractall(ontologyName(_)),
	retractall(ontology(_,_)),
	%assert(ontologyName(File)),
	((open(File, read, O), ON=File); (atom_concat(File,'.pl',ON), open(ON,read,O))),
	repeat,
	  read(O, Clause),
	  (Clause=(A:-B) -> assertz(ontology(ON,A,B))
	                  ; assertz(ontology(ON,A,true))),
	  at_end_of_stream(O),
	!,
	close(O).

%%% Ontology->Unload: unloads life-cycle ontology from knowledge base
unloadOntology(O) :-
%	retract(ontologyName(O)),
	retractall(ontology(O,_)).

%test_service(uri1, stephen, nn(k(A,B),k(C,D))=>[instance_abstraction_level(A),numeric_value_pairs(B),exclusive_domain_abstraction_level(C),associationist_represenation(D),generalisation_synthesis(k(A,B),k(C,D))]).

test_service(uri1,lifecycle('stephen.pl', neural_network_training(k(A,B),k(C,D))=>[instance_abstraction_level(A),numeric_value_pairs(B),exclusive_domain_abstraction_level(C),associationist_represenation(D),generalisation_synthesis(k(A,B),k(C,D))])).

test_service(uri2,lifecycle('stephen.pl', neural_network_preprocessing(k(A,B),k(C,D))=>[instance_abstraction_level(A),symbolic_value_pairs(B),instance_abstraction_level(C),numeric_value_pairs(D),syntactic_translation(k(A,B),k(C,D))])).

test_query(uri3,lifecycle('stephen.pl', query(k(A,B),k(C,D))=>[instance_abstraction_level(A),symbolic_value_pairs(B),exclusive_domain_abstraction_level(C),formal(D),generalisation_synthesis(k(A,B),k(C,D))])).



test_service(query,query(uri2, ont, nnq(k(A,B),k(C,D))=>[instance_abstraction_level(A),numeric_value_pairs(B),exclusive_domain_abstraction_level(C),associationist_represenation(D),generalisation_synthesis(k(A,B),k(C,D))])).

go:-
   retractall(sevice(_,_,_,_,_)),
   retractall(ontology(_,_)),
   assert_services.


assert_services :-
   test_service(URI,lifecycle(Ont,Lifecycle)),
   get_transformation(URI,Ont,Lifecycle,Name,Trans),
   assert(service(URI,Name,Ont,Lifecycle,Trans)),
   format('Asserted ~q', [service(URI,Name,Ont,Lifecycle,Trans)]),
   fail.
assert_services.


%%%
%%% extracts a transformation from URI, Ontology and Lifecycle...
%%%
get_transformation(URI,Ontology,Lifecycle,Name,Transformation):-
   unloadOntology(Ontology),
   loadOntology(Ontology),
   break_lifecycle(Ontology,Lifecycle,Name,Transformation).


%%%
%%% extracts from the lifecycle (X=>...) the transformation.
%%%
break_lifecycle(Ontology,X=>[Term|Terms], Name, transformation(Input,Output,Trans)):-
   X=..[Name,In,Out],
   get_know(Ontology,input_knowledge,In, [Term], Terms, Input),
   remove_terms(Input,[Term|Terms],[R|Rest]),
   get_know(Ontology,output_knowledge,Out,[R],Rest, Output),
   remove_terms(Output,[R|Rest],Trans).


remove_terms([],Terms,Terms).
remove_terms([I|In],Terms,Rest):-
   select(I,Terms,O),
   remove_terms(In,O,Rest).

   
get_know(Ontology,KT, Know, Head, [Term|Terms],Result):-
   prove_k(Ontology, KT, Know, Head, Result).
get_know(Ontology, KT, Know, Head,[Term|Terms], Result):-
   get_know(Ontology, KT, Know, [Term|Head], Terms, Result).
get_know(_Ontology,_KT,_Know,_,[],[]).

prove_k(Ontology, KT,In,Terms,Proof):-
% can Terms be derived from In?
   F=..[KT,In],
   make_clause(F,Terms,TermToProve),
   satisfy_query(Ontology, TermToProve,Proof).
   %format('result is ~q', [Terms]),


make_clause(In,Terms,(In:-RHS)):-
   make_cl(Terms,RHS).
make_cl([Term],Term).
make_cl([T|Terms],Ret):-
   Ret = (T,Rest),
   make_cl(Terms,Rest).


satisfy_query(Ontology,(A :- B),P) :-
    believe(Ontology,B),
    ( (satisfied(Ontology,A,P), !, forget) ; (forget, !, fail) ).
%    ( (satisfied(A), forget) ; (forget, fail) ).
satisfy_query(Ontology,Goal,P) :-
    \+ Goal = (_ :- _),
    satisfied(Ontology,Goal,P).

forget :-
    retractall(believed(_,_)).

believe(Ontology,(A,B)) :-
    believe(Ontology,A),
    believe(Ontology,B).
believe(Ontology,X) :-
    \+ X = (_,_),
    add_belief(Ontology,X).

add_belief(Ontology,Fact) :-
    \+ believed(Ontology,Fact),
    assert(believed(Ontology,Fact)), !.
add_belief(_,_).

satisfied(Ontology,true,[]):- !.
satisfied(Ontology,(A, B),P):- !,
    satisfied(Ontology,A,P1),
    satisfied(Ontology,B,P2),
    append(P1,P2,P).
satisfied(Ontology,(A ; B),P):- !,
    satisfied(Ontology,A,P) ; satisfied(Ontology,B,P).
satisfied(Ontology,Fact,[Fact]):- 
    believed(Ontology,Fact).

%satisfied(Meta):-
%    meta_pred(Meta, Call), !,
%    Call.

%satisfied(Ass):-
%    assert_pred(Ass, Call), !,
%    Call.
%satisfied(Ret):-
%    retract_pred(Ret, Call), !,
%    Call.

%satisfied(Goal):-
%    predicate_property(Goal, built_in), !,
%    Goal.
%satisfied(Goal):-
%    predicate_property(Goal, compiled), !,
%    Goal.
%satisfied(Goal):-
%    predicate_property(Goal, interpreted), !,
%    Goal.

satisfied(Ontology,A,P):-
    ontology(Ontology, A, B),
    satisfied(Ontology, B, P).

%%%
%%% (Dave's!) Original meta-interpreter:
%%%

%satisfied(true):- !.
%satisfied((A, B)):- !,
%    satisfied(A),
%    satisfied(B).
%satisfied((A ; B)):- !,
%    satisfied(A) ; satisfied(B).
%satisfied(Fact):- 
%    believed(Fact).
%satisfied(Meta):-
%    meta_pred(Meta, Call), !,
%    Call.
%satisfied(Ass):-
%    assert_pred(Ass, Call), !,
%    Call.
%satisfied(Ret):-
%    retract_pred(Ret, Call), !,
%    Call.
%satisfied(Goal):-
%    predicate_property(Goal, built_in), !,
%    Goal.
%satisfied(Goal):-
%    predicate_property(Goal, compiled), !,
%    Goal.
%satisfied(Goal):-
%    predicate_property(Goal, interpreted), !,
%    Goal.
%satisfied(A):-
%    ontology(_File, A, B),
%    satisfied(B).


