%%%
%%% stephen potter july '03
%%% a first try at a generic 'transformation' broker
%%% 

%%%
%%% The brokering algorithm is as follows:
%%% 1. Let the solution set, S={}. Let the possibles set, P={}. Search
%%% the advertised services to find possible 'terminal' services for
%%% the current query (services that *may* constitute the final element 
%%% in a sequence of services). Add each of these to P to represent the
%%% final element of a potential service sequence.
%%% 
%%% 2. If P={} then stop. Return S as the solution to this query.
%%%
%%% 3. Consider each element p of P in turn to determine whether it 
%%% constitutes a 'complete' service sequence for the query: if the
%%% input knowledge of the query and of the first element of p match
%%% then this is a complete sequence and p is transferred from P to S.
%%%
%%% 4. For each incomplete sequence p in P search the current services
%%% for any other known service s that could plausibly precede these in
%%% a service sequence: s can plausibly precede p is the output
%%% knowledge of s matches the input knowledge of p. For each s that 
%%% meets this criterion, a new plausible service sequence consisting
%%% of s appended to the front of p is added to P. Finally, remove p
%%% from P.
%%%
%%% 5. Repeat from step 2.

%%%
%%% Representations:
%%%
%%% A service is represented by the tuple:
%%%   service(URI, Name, Ontology, Lifecycle, Transformation)
%%% where
%%%   * URI is the URI of the service in question.
%%%   * Name is the name of this service.
%%%   * Ontology is the (name of the prolog file containing the) transformation 
%%% ontology used to describe the Transformation.
%%%   * Lifecycle is the lifecycle itself, in the form:
%%%      trans(input,output)=>[terms+]
%%%      if input, output are not ground, it is expected that the variables
%%%   will appear in the (Prolog) terms (effectively used to define the 
%%%   constraints on input and output and the transformation from one to the
%%%   other). Note that terms that refer (perhaps indirectly) to input concern
%%%   the input knowledge, those refering to output concern the output 
%%%   knowledge, and those refering to both concern the transformation.
%%%   * Transformation is the transformation implied within the lifecycle: this
%%%   is the tuple transformation(input,output,transformation). Each of the 
%%%   elements input and output is a list of lists, each of which represents the
%%%   description of one body of knowledge. 
%%% 
%%% A service query will have the same form. 
:- use_module(library(lists)).
:- op(800, xfx, [=>]).
:- dynamic ontology/3, believed/2, service/5, request/5, serviceList/1.

%%% Ontology->Load: loads life-cycle ontology from file
loadOntology(File) :-
	%retractall(ontologyName(_)),
	retractall(ontology(File,_,_)),
	%assert(ontologyName(File)),
	%((open(File, read, O), ON=File); (atom_concat(File,'.pl',ON), open(ON,read,O))),
        open(File,read,O),
        %retractall(ontology(ON,_,_)),
	repeat,
	  read(O, Clause),
          ((\+Clause=end_of_file,
	  (Clause=(A:-B) -> assertz(ontology(File,A,B))
	                  ; assertz(ontology(File,A,true))));true),
	  at_end_of_stream(O),
	!,
	close(O).

%%% Ontology->Unload: unloads life-cycle ontology from knowledge base
unloadOntology(O) :-
%	retract(ontologyName(O)),
	retractall(ontology(O,_,_)).

%test_service(uri1, stephen, nn(k(A,B),k(C,D))=>[instance_abstraction_level(A),numeric_value_pairs(B),exclusive_domain_abstraction_level(C),associationist_representation(D),generalisation_synthesis(k(A,B),k(C,D))]).

test_service(uri1,lifecycle('stephen.pl', neural_network_training(k(A,B),k(C,D))=>[instance_abstraction_level(A),numeric_value_pairs(B),exclusive_domain_abstraction_level(C),associationist_representation(D),generalisation_synthesis(k(A,B),k(C,D))])).

test_service(uri2,lifecycle('stephen.pl', neural_network_preprocessing(k(A,B),k(C,D))=>[instance_abstraction_level(A),symbolic_value_pairs(B),instance_abstraction_level(C),numeric_value_pairs(D),syntactic_translation(k(A,B),k(C,D))])).

test_query(uri3,lifecycle('stephen.pl', query(k(A,B),k(C,D))=>[instance_abstraction_level(A),symbolic_value_pairs(B),exclusive_domain_abstraction_level(C),formal(D),generalisation_synthesis(k(A,B),k(C,D))])).



test_service(query,query(uri2, ont, nnq(k(A,B),k(C,D))=>[instance_abstraction_level(A),numeric_value_pairs(B),exclusive_domain_abstraction_level(C),associationist_representation(D),generalisation_synthesis(k(A,B),k(C,D))])).

go:-
   retractall(sevice(_,_,_,_,_)),
   retractall(request(_,_,_,_,_)),
   retractall(ontology(_,_,_)),
   retractall(serviceList(_)),
   %assert(serviceList([])),
   assert_services([uri1,uri2],SL),
   assert(serviceList(SL)),
   assert_request,
   broker.


assert_services([URI|Us],[service(URI,Name,Ont,Lifecycle,Trans)|R]) :-
   test_service(URI,lifecycle(Ont,Lifecycle)),
   get_transformation(URI,Ont,Lifecycle,Name,Trans),
   %serviceList(OldSL),
%   retractall(serviceList(_)),
%   append([service(URI,Name,Ont,Lifecycle,Trans)],OldSL,NewSL),
 %  assert(serviceList(NewSL)),
   %format('Asserted ~q', [service(URI,Name,Ont,Lifecycle,Trans)]),
   assert_services(Us,R).
assert_services([],[]).

assert_request:-
   test_query(URI,lifecycle(Ont,Lifecycle)), %Lifecycle may be partial
   get_transformation(URI,Ont,Lifecycle,Name,Trans),
   assert(request(URI,Name,Ont,Lifecycle,Trans)).
  % format('Asserted ~q', [request(URI,Name,Ont,Lifecycle,Trans)]).


%%%
%%% extracts a transformation from URI, Ontology and Lifecycle...
%%%
get_transformation(URI,Ontology,Lifecycle,Name,Transformation):-
   unloadOntology(Ontology),
   loadOntology(Ontology),
   break_lifecycle(Ontology,Lifecycle,Name,Transformation).


%%%
%%% extracts from the lifecycle (X=>...) the transformation.
%%%
break_lifecycle(Ontology,X=>[Term|Terms], Name, transformation(Input,Output,Trans)):-
   X=..[Name,In,Out],
   get_know(Ontology,input_knowledge,In, [Term], Terms, Input),
   remove_terms(Input,[Term|Terms],[R|Rest]),
   get_know(Ontology,output_knowledge,Out,[R],Rest, Output),
   remove_terms(Output,[R|Rest],Trans).


remove_terms([],Terms,Terms).
remove_terms([I|In],Terms,Rest):-
   select(I,Terms,O),
   remove_terms(In,O,Rest).

   
get_know(Ontology,KT, Know, Head, [Term|Terms],Result):-
   prove_k(Ontology, KT, Know, Head, Result).
get_know(Ontology, KT, Know, Head,[Term|Terms], Result):-
   get_know(Ontology, KT, Know, [Term|Head], Terms, Result).
get_know(_Ontology,_KT,_Know,_,[],[]).

prove_k(Ontology, KT,In,Terms,Proof):-
% can Terms be derived from In?
   F=..[KT,In],
   make_clause(F,Terms,TermToProve),
   satisfy_query(Ontology, TermToProve,Proof).
   %format('result is ~q', [Terms]),


make_clause(In,Terms,(In:-RHS)):-
   make_cl(Terms,RHS).
make_cl([Term],Term).
make_cl([T|Terms],Ret):-
   Ret = (T,Rest),
   make_cl(Terms,Rest).


satisfy_query(Ontology,(A :- B),P) :-
    believe(Ontology,B),
    ( (satisfied(Ontology,A,P), !, forget) ; (forget, !, fail) ).
satisfy_query(Ontology,Goal,P) :-
    \+ Goal = (_ :- _),
    satisfied(Ontology,Goal,P).

forget :-
    retractall(believed(_,_)).

believe(Ontology,(A,B)) :-
    believe(Ontology,A),
    believe(Ontology,B).
believe(Ontology,X) :-
    \+ X = (_,_),
    add_belief(Ontology,X).

add_belief(Ontology,Fact) :-
    \+ believed(Ontology,Fact),
    assert(believed(Ontology,Fact)), !.
add_belief(_,_).

satisfied(Ontology,true,[]):- !.
satisfied(Ontology,(A, B),P):- !,
    satisfied(Ontology,A,P1),
    satisfied(Ontology,B,P2),
    append(P1,P2,P).
satisfied(Ontology,(A ; B),P):- !,
    satisfied(Ontology,A,P) ; satisfied(Ontology,B,P).
satisfied(Ontology,Fact,[Fact]):- 
    believed(Ontology,Fact).
satisfied(Ontology,A,P):-
    ontology(Ontology, A, B),
    satisfied(Ontology, B, P).
%satisfied(Meta):-
%    meta_pred(Meta, Call), !,
%    Call.
%satisfied(Ass):-
%    assert_pred(Ass, Call), !,
%    Call.
%satisfied(Ret):-
%    retract_pred(Ret, Call), !,
%    Call.
%satisfied(Goal):-
%    predicate_property(Goal, built_in), !,
%    Goal.
%satisfied(Goal):-
%    predicate_property(Goal, compiled), !,
%    Goal.
%satisfied(Goal):-
%    predicate_property(Goal, interpreted), !,
%    Goal.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% actual broker stuff:
%%%
%%%

ssubsumes(_O,T,T).
ssubsumes(Ontology,T1,T2):-
   %true if T1 subsumes T2
   ontology(Ontology,T1,T2).
ssubsumes(Ontology,T1,T2):-
   ontology(Ontology,T3,T2),
   ssubsumes(Ontology,T1,T3).

%%% D represents the 'distance' between the terms in the ontology.
ssubsumesd(_O,T,T,0).
ssubsumesd(Ontology,T1,T2,1):-
   %true if T1 subsumes T2
   ontology(Ontology,T1,T2).
ssubsumesd(Ontology,T1,T2,D):-
   ontology(Ontology,T3,T2),
   ssubsumesd(Ontology,T1,T3,D1),
   D is D1+1.


broker:-
  serviceList(SL),
  find_possible_terminals(SL,P),
  format('Possibles: ~q',[P]),
  check_terminals(P,PT),
  ((broker_stop_condition(PT),PT=Ans);
   extend_service_chains(PT,Ans)),
   format('Answer: ~q',[Ans]).

%%%
%%% 
%%%     chain(I,S,Terminates,Reqs)
%%% where I is a list of instances (ie of instance(Type,Name,Values)) of
%%% competences, such that the output of the last of the list is compatible
%%% with the desired output (or desired transformation-type, if no output is
%%% specified) of the query. The intention is to add to this a
%%% list of other instances such that the first in the list matches the  
%%% input knowledge of the query, and each intermediate instance transforms
%%% its inputs into the outputs of the next, culminating in the final, 
%%% desired transformation.
%%% S stores an intermediate 'score' of the current list.
%%% Terminates takes a value from {yes,no}: yes indicates that the chain
%%% represents a valid transformation compatible with the query; no indicates
%%% that the chain is not yet complete (but is still compatible - incompatible
%%% chains are removed from the list).
%%% Reqs is a list of the additional knowledge requirements that the chain
%%% would entail (ie since any transformation can have more than one input,
%%% additional input knowledge might be necessary at point in the chain).
%%%

find_possible_terminals([],[]).
find_possible_terminals([Service|SL],[chain([Service],Score,no,[])|PR]):-
   % note: have to have same ontology!
   Service = service(URI,Name,Ont,_Lc,transformation(_In,Out,Trans)),
   request(_RU,_RNm,Ont,_RLc,transformation(_RIn,ROut,RTrans)),
   % match output knowledge:
   match_function(Ont,Out,ROut,OutScore),
   % match transformation:
   match_function(Ont,Trans,RTrans,TransScore),
   Score is OutScore+TransScore,
   % needs to be sufficiently similar:
   Score < 10000,
   find_possible_terminals(SL,PR).
find_possible_terminals([_|SL],P):-
   find_possible_terminals(SL,P).


%%%
%%% So: the service (input/output) knowledge will be a list of terms, each of
%%% which must be subsumed by a term of the query (input/output) knowledge;
%%% however, it is not known which (if any) term this will be... 
match_function(_,[],_,0).
match_function(_,_,[],0).
match_function(Ont,SK,RK,Score):-
	generate_kscore_matrix(Ont,SK,RK,Matrix),
	find_min_coords(Matrix,X,Y,Min),
	nth(Y,SK,ST),	
	select(ST,SK,NSK),
	nth(X,RK,RT),
	select(RT,RK,NRK),
	match_function(Ont,NSK,NRK,S2),
	Score is Min + S2.

%%%
%%% generate_iscore_matrix(+L1,+L2,-ScoreMatrix) generates a list of
%%% lists of the scores generated by matching each item in L1 with 
%%% each item in L2, through repeated calls to match_instance_values.
%%%
generate_kscore_matrix(_Ont,_,[],[]).
generate_kscore_matrix(Ont,SK,[RK|RRK],[SL|MR]):-
	match_kterm_values(Ont,SK,RK,SL),
	generate_kscore_matrix(Ont,SK,RRK,MR).

%%%
%%% match_kterm_values(Ont,+L1,+I,-ScoreL) matches I with each instance
%%% in list L1, recording the corresponding scores in the list ScoreL.
%%%
match_kterm_values(_Ont,[],_,[]).
match_kterm_values(Ont,[C|AC],Q,[S1|S]):-
	ssubsumesd(Ont,Q,C,S1),
	!,
	match_kterm_values(Ont,AC,Q,S).
% or, if there is no match:
match_kterm_values(Ont,[_C|AC],Q,[10000|S]):-
	!,
	match_kterm_values(Ont,AC,Q,S).


%%%
%%% find_min_coords(+Matrix,-X,-Y,-Min): given a matrix (list of lists) of
%%% values, finds & returns the coordinates X and Y of the minimum value
%%% Min.   
%%%
find_min_coords(Matrix,X,Y,Min):-
	generate_min_list(Matrix,MinList),
	min_list(MinList,Min),
	min_position(MinList,Min,X),
	nth(X,Matrix,YList),
	min_position(YList,Min,Y).

%%%
%%% generate_min_list(+Matrix,-MinList) takes a list of lists of integers,
%%% and via calls to min_list, returns a list of the min value in each
%%% list.
%%% 
generate_min_list([],[]).
generate_min_list([M1|MR],[Min|MinL]):-
	min_list(M1,Min),
	generate_min_list(MR,MinL).

%%%
%%% min_position(+ML,+Min,-Pos) returns Pos, the first occurrence of
%%% the value Min in the list ML.
%%%
min_position([Min|_],Min,1).
min_position([M|ML],Min,S):-
	\+M = Min,
	min_position(ML,Min,S1),
	S is S1 + 1.



%%%%
%%%%
%%%%

check_terminals([],[]).
check_terminals([chain([S|SL],Score,no,Req)|Chains],[chain([S|SL],NewScore,yes,Req)|CChains]):-
   S = service(URI,Name,Ont,_Lc,transformation(In,_Out,_Trans)),
   request(_RU,_RNm,Ont,_RLc,transformation(RIn,_ROut,_RTrans)),
   % match input knowledge:
   match_function(Ont,In,RIn,InScore),
   NewScore is InScore+Score,
   % needs to be sufficiently similar:
   NewScore < 10000,
   check_terminals(Chains,CChains).
check_terminals([C|Cs],[C|Ss]):-
   check_terminals(Cs,Ss).   

%%%
%%% broker_stop_condition(I) is true if all chains in I terminate ('yes'),
%%% or else if the list of chains is empty.
%%%
broker_stop_condition([]).
broker_stop_condition([chain(_,_,yes,_)|Cs]):-
	broker_stop_condition(Cs).
broker_stop_condition([chain(_,_,no,_)|_]):-
	!,fail.


extend_service_chains(PL,Ans):-
       extend_chains(PL,LL),
       check_terminals(LL,P1),
       ((broker_stop_condition(P1),Ans=P1);extend_service_chains(P1,Ans)).



%%%
%%% extend_chains(InCL,OutCL) attempts to extend the non-terminated
%%% chains in InCL through a call to find_plink; successful
%%% extensions are recorded in OutCL, as are any terminated chains.
%%%
extend_chains([],[]).
extend_chains([chain(I,S,yes,R)|Is],[chain(I,S,yes,R)|LL]):-
	extend_chains(Is,LL).
extend_chains([chain(I,S,no,R)|Is],LL):-
	serviceList(SL),
	find_service_link(chain(I,S,no,R),SL,RL),
	extend_chains(Is,TL),
	append(RL,TL,LL).

%%%
%%% find_service_link(Chain,Comps,OutCL) tries to extend the chain Chain using
%%% each competence in Comps in turn: if the input-knowledge for the first 
%%% element in the chain is not incompatible with the output-knowledge of
%%% the competence in question (determined via a call to match_link), then
%%% adding the competence to the head of the chain, and the chain to the
%%% new chain list OutCL.  
%%% (A check is made to ensure that the competence in question does not
%%% already occur in the chain in order to avoid cycles....)
%%% (Another check is added: for incompatible services: if two services have
%%% been already found to be incompatible (perhaps during an earlier query),
%%% there is no need to check whether their composition represents a valid
%%% extension of the current sequence. This is added for reasons of
%%% efficiency.) 
%%%
find_service_link(_,[],[]).
%find_service_link(chain([service(_U,_N,Ont,_L,transformation(In,Out,Trans)|Ss],Sc,no,Reqd),[service(_U1,_N1,Ont,_L,transformation(In1,Out1,Trans1)|R],LL):-
	%incompatible(N,N1),
%	format("~w and ~w are incompatible.\n",[N,N1]),
%	find_service_link(chain([instance(T,N,V)|Is],Sc,no,Reqd),R,LL).
find_service_link(chain(SC,Sc,no,Reqd),[S|R],LL):-
	%already_in_chain([instance(T,N,V)|Is],N1),
        member(S,SC),
	find_service_link(chain(SC,Sc,no,Reqd),R,LL).
find_service_link(chain(Chain,Sc,no,Reqd),[Service|R],LL):-
        Chain = [service(_U,_N,Ont,_L,transformation(In,_Out,_Trans))|Ss],
        Service = service(_U1,_N1,Ont,_L1,transformation(_In1,Out1,_Trans1)),
	match_function(Ont,Out1,In,Score),
        TotalScore is Sc + Score,
        TotalScore < 10000,
        append([Service],Chain,NewChain),
	find_service_link(chain(Chain,Sc,no,Reqd),R,RL),
	append([chain(NewChain,TotalScore,no,Reqd)],RL,LL).


%incompatible(N,N1):-
%	incompatibles(N,IL),
%	member(N1,IL).

%add_incompatible(N,N1):-
%	incompatibles(N,OIL),
%	retractall(incompatibles(N,_)),
%	append([N1],OIL,NIL),
%	format("~w and ~w have been found to be incompatible.\nIncompatibles:~w\n",[N,N1,NIL]),
%	assert(incompatibles(N,NIL)).

%extend_chain(chain(I,Sc,_,Reqd),NI,S,[chain(RI,NS,no,Reqd)]):-
%	S<10000,
%	append([NI],I,RI),
%	NS is Sc + S.
%extend_chain(chain([instance(_T,N,_V)|Is],_,_,_),instance(_T1,N1,_V1),_,[]):-
%	add_incompatible(N,N1).

%already_in_chain([],_) :-
%	!, fail.
%already_in_chain([service(_,Name,_)|_],Name).
%already_in_chain([_|Is],Name):-
%	already_in_chain(Is,Name).
